/*
 * ramblock.c - A simple RAM based block device.
 *
 * Author: Jon Masters for Linux User & Developer
 *
 * Original idea from online examples and LWN.net 2.6 driver-porting series.
 *
 * This program is distributed under the GPL.
 *
 */

#include <linux/module.h>         /* module macros */
#include <linux/init.h>           /* init macros */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/kernel.h>

#include <linux/types.h>          /* for custom kernel types */
#include <linux/fs.h>             /* file_operations struct */
#include <linux/fcntl.h>          /* file seeking */
#include <linux/cdev.h>           /* cdev functions */
#include <linux/vmalloc.h>        /* vmalloc functions */
#include <linux/genhd.h>          /* gendisk functions */
#include <linux/blkdev.h>         /* block functions */
#include <linux/hdreg.h>          /* disk geometry */
#include <asm/uaccess.h>          /* userspace access */

#define RAMBLOCK_SIZE (1024*1024) /* 1M storage */

typedef struct ramblock_struct {

  unsigned long size;             /* device size */
  u8 *data;                       /* device data */
  
  dev_t major;                    /* major number */
  spinlock_t lock;                /* spinlock for low-level routines */
  struct request_queue *queue;    /* request queue for device */
  struct gendisk *disk;           /* generic disk */

} ramblock_t;

int ramblock_ioctl(struct inode *inode, struct file *filp,
		   unsigned int cmd, unsigned long arg);

static struct block_device_operations ramblock_fops = {
  .owner = THIS_MODULE,
  .ioctl = ramblock_ioctl
};

static ramblock_t ramblock_dev;  /* the "disk" */

const char *ramblock_name = "ramblock";           /* for reference */

#define KERNEL_SECTOR_SIZE 512                    /* kernel deals in sectors */

static void ramblock_request(request_queue_t *queue)
{

  struct request *current_request;

  unsigned long offset = 0;           /* offset in ramblock_dev.data */
  unsigned long size = 0;             /* size of the request */
  
  while ((current_request = elv_next_request(queue)) != NULL) {
    if ( !blk_fs_request(current_request)) {
      printk(KERN_NOTICE "%s: not a valid request for this device.\n",
	     ramblock_name);
      end_request(current_request,0);
      continue;
    }

    offset = current_request->sector*KERNEL_SECTOR_SIZE;
    size = current_request->current_nr_sectors*KERNEL_SECTOR_SIZE;

    if ((offset + size) > ramblock_dev.size) {
      printk(KERN_NOTICE "%s: Access beyond end of device.\n",
	     ramblock_name);
      goto out;
      return;
    }

    if (rq_data_dir(current_request)) { /* write */
      memcpy(ramblock_dev.data + offset, current_request->buffer, size);
    } else {
      memcpy(current_request->buffer, ramblock_dev.data + offset, size);
    }
    
out:
    end_request(current_request, 1); /* always ok for our simple device */
  }
}

int ramblock_ioctl(struct inode *inode, struct file *filp,
		   unsigned int cmd, unsigned long arg)
{

  /* Just need to implement HDIO_GETGEO for partitioning support */

  struct hd_geometry geo;

  switch(cmd) {

  case HDIO_GETGEO:
    geo.cylinders = (ramblock_dev.size >> 4);
    geo.heads = 4;
    geo.sectors = 16;
    geo.start = 4;
    
    if (copy_to_user((void *)arg, &geo, sizeof(geo))) {
      return -EFAULT;
    }
    
    return 0;
    
  }
  return -ENOTTY; /* unknown command */
}

int __init ramblock_init(void)
{

	/* Init ramblock_dev */
	
	ramblock_dev.size = RAMBLOCK_SIZE;
	if (!(ramblock_dev.data = vmalloc(ramblock_dev.size))) {
		return -ENOMEM;
	}

	memset(ramblock_dev.data, '\0', ramblock_dev.size); /* zero data */
	
	/* Init spinlock */
	
	spin_lock_init(&ramblock_dev.lock);

	/* Init request queue */

	if ((ramblock_dev.queue = blk_init_queue(ramblock_request, &ramblock_dev.lock)) == 0) {
		printk(KERN_WARNING "%s: unable to allocate queue.\n",
				ramblock_name);
		goto error;
	}

	/* Register blkdev
	 * Request an automatic major number assignment
	 */

	if ((ramblock_dev.major = register_blkdev(0, ramblock_name)) <= 0) {
		printk(KERN_WARNING "%s: unable to register blkdev.\n",
				ramblock_name);
		goto error;
	}

	/* Register gendisk */

	if (NULL == (ramblock_dev.disk = alloc_disk(16))) { /* 16 partitions */
		printk(KERN_WARNING "%s: unable to register gendisk.\n",
				ramblock_name);
		unregister_blkdev(ramblock_dev.major, ramblock_name);
		goto error;
	}

	/* Fill in ramblock_dev.disk fields */

	ramblock_dev.disk->major = ramblock_dev.major;
	ramblock_dev.disk->first_minor = 0;
	ramblock_dev.disk->fops = &ramblock_fops;
	ramblock_dev.disk->private_data = &ramblock_dev;
	ramblock_dev.disk->queue = ramblock_dev.queue;
	
	strcpy(ramblock_dev.disk->disk_name, ramblock_name);
	set_capacity(ramblock_dev.disk, ramblock_dev.size/KERNEL_SECTOR_SIZE);
	add_disk(ramblock_dev.disk);

	/* Print a banner message */
	
	printk(KERN_INFO "%s: registered block device (at 0x%08x, size %d) on major %d.\n",
	       ramblock_name,
	       (unsigned int)(ramblock_dev.data),
	       (unsigned int)ramblock_dev.size,
	       (unsigned int)(ramblock_dev.major));
	return 0;
	
   error:
   	vfree(ramblock_dev.data);
	return -ENOMEM;
}

static void __exit ramblock_exit(void)
{
	printk(KERN_INFO "%s: unregistering block device %d",
			ramblock_name, ramblock_dev.major);

	/* Remove disk from system */

	del_gendisk(ramblock_dev.disk);

	/* Tell others to leave it alone */

	put_disk(ramblock_dev.disk);

	/* Remove the block device */

	unregister_blkdev(ramblock_dev.major, ramblock_name);

	/* Empty final queue contents */

	blk_cleanup_queue(ramblock_dev.queue);

	/* Free up ramblock memory */

	vfree(ramblock_dev.data);
}

module_init(ramblock_init);
module_exit(ramblock_exit);

MODULE_AUTHOR("Jon Masters <jcm@jonmasters.org>");
MODULE_LICENSE("GPL");
